package core

import (
	"bytes"
	"context"
	"fmt"
	"net/url"
	"text/template"
	"time"

	"github.com/fluxcd/pkg/apis/meta"
	"github.com/fluxcd/pkg/runtime/conditions"
	"github.com/minio/minio-go/v7"
	corev1 "k8s.io/api/core/v1"
	networkingv1 "k8s.io/api/networking/v1"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/util/rand"
	configv1alpha1 "libre.sh/controller/apis/config/v1alpha1"
	corev1alpha1 "libre.sh/controller/apis/core/v1alpha1"
	"libre.sh/controller/internal"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/controller/controllerutil"
	"sigs.k8s.io/controller-runtime/pkg/log"
)

const ownerPolicy = `
{
	"Version": "2012-10-17",
	"Statement": [
	  {
		"Effect": "Allow",
		"Action": ["s3:*"],
		"Resource": [
		  "arn:aws:s3:::${aws:username}/*",
		  "arn:aws:s3:::${aws:username}"
		]
	  }
	]
}
`

const cdnPolicy = `
{
	"Version": "2012-10-17",
	"Statement": [
	  {
		"Action": ["s3:GetBucketLocation"],
		"Effect": "Allow",
		"Principal": { "AWS": ["*"] },
		"Resource": ["arn:aws:s3:::{{ . }}"]
	  },
	  {
		"Action": ["s3:GetObject"],
		"Effect": "Allow",
		"Principal": { "AWS": ["*"] },
		"Resource": ["arn:aws:s3:::{{ . }}/*","arn:aws:s3:::{{ . }}/**"]
	  }
	]
  }  
`

const bucketFinalizerName = "core.libre.sh/finalizer"

// BucketReconciler reconciles a Bucket object
type BucketReconciler struct {
	client.Client
	Scheme *runtime.Scheme
	Config *configv1alpha1.LibreConfig
}

//+kubebuilder:rbac:groups=core.libre.sh,resources=buckets,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=core.libre.sh,resources=buckets/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=core.libre.sh,resources=buckets/finalizers,verbs=update

// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the Bucket object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.12.1/pkg/reconcile
func (r *BucketReconciler) Reconcile(ctx context.Context, req ctrl.Request) (result ctrl.Result, err error) {
	log := log.FromContext(ctx)

	var bucket corev1alpha1.Bucket
	err = r.Client.Get(ctx, req.NamespacedName, &bucket)
	if err != nil {
		return ctrl.Result{}, client.IgnoreNotFound(err)
	}

	if bucket.ObjectMeta.DeletionTimestamp.IsZero() {
		if !controllerutil.ContainsFinalizer(&bucket, bucketFinalizerName) {
			controllerutil.AddFinalizer(&bucket, bucketFinalizerName)
			if err := r.Update(ctx, &bucket); err != nil {
				return ctrl.Result{}, err
			}
		}
	} else {
		if controllerutil.ContainsFinalizer(&bucket, bucketFinalizerName) {
			if bucket.Spec.Purge {
				if err := r.deleteExternalResources(ctx, &bucket); err != nil {
					return ctrl.Result{}, err
				}
			}

			controllerutil.RemoveFinalizer(&bucket, bucketFinalizerName)
			if err := r.Update(ctx, &bucket); err != nil {
				return ctrl.Result{}, err
			}
		}
		return ctrl.Result{}, nil
	}

	if bucket.Spec.Suspend {
		return ctrl.Result{}, nil
	}

	log.Info("Reconciling")

	var minioErr error

	defer func() {
		if err == nil && minio.IsNetworkOrHostDown(minioErr, false) {
			err = minioErr
		}
		if err != nil {
			return
		}
		if minioErr != nil {
			log.Error(minioErr, "Failed to provision bucket")
			result.RequeueAfter = 10 * time.Minute
			conditions.MarkStalled(&bucket, meta.FailedReason, minioErr.Error())
		} else {
			conditions.Delete(&bucket, meta.StalledCondition)
		}
		if conditions.IsStalled(&bucket) {
			conditions.MarkFalse(&bucket, meta.ReadyCondition, meta.FailedReason, "Is stalled")
		} else if conditions.IsReconciling(&bucket) {
			conditions.MarkUnknown(&bucket, meta.ReadyCondition, meta.ProgressingReason, "Is reconciling")
		} else {
			conditions.MarkTrue(&bucket, meta.ReadyCondition, meta.SucceededReason, "Is provisionned")
		}
		err = r.Client.Status().Update(ctx, &bucket)
	}()

	if bucket.Generation != conditions.GetObservedGeneration(&bucket, meta.ReadyCondition) {
		conditions.MarkReconciling(&bucket, meta.ProgressingReason, "Reconciling in progress")
		err = r.Client.Status().Update(ctx, &bucket)
		if err != nil {
			return
		}
	}

	mc, minioErr := internal.GetMinioClient(ctx, r.Client, r.Config, &bucket)
	if minioErr != nil {
		return
	}

	buckerExits, minioErr := mc.BucketExists(ctx, bucket.Spec.BucketName)
	if minioErr != nil {
		return
	}
	if !buckerExits {
		// TODO region
		minioErr = mc.MakeBucket(ctx, bucket.Spec.BucketName, minio.MakeBucketOptions{})
		if minioErr != nil {
			return
		}

	}

	bucketEndpoint := mc.Provider.Endpoint
	if mc.Provider.BucketDNS != nil {
		bucketEndpoint = fmt.Sprintf("%s.%s", bucket.Spec.BucketName, mc.Provider.Endpoint)
	}

	var secret corev1.Secret
	secret.Name = bucket.GetBucketSecretName()
	secret.Namespace = bucket.Namespace
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &secret, func() error {
		secretKey := secret.Data["secretKey"]
		if len(secretKey) == 0 {
			secretKey = []byte(rand.String(32))
		}
		var url url.URL
		url.Host = mc.Provider.Endpoint
		if mc.Provider.Insecure {
			url.Scheme = "http"
		} else {
			url.Scheme = "https"
		}
		secret.Data = map[string][]byte{
			"accessKey": []byte(bucket.Spec.BucketName),
			"secretKey": secretKey,
			"endpoint":  []byte(bucketEndpoint),
			"url":       []byte(url.String()),
			"bucket":    []byte(bucket.Spec.BucketName),
		}
		return controllerutil.SetOwnerReference(&bucket, &secret, r.Scheme)
	})
	if err != nil {
		return
	}

	minioErr = mc.Admin.AddUser(ctx, bucket.Spec.BucketName, string(secret.Data["secretKey"]))
	if minioErr != nil {
		return
	}

	minioErr = mc.Admin.AddCannedPolicy(ctx, "owner", []byte(ownerPolicy))
	if minioErr != nil {
		return
	}

	minioErr = mc.Admin.SetPolicy(ctx, "owner", bucket.Spec.BucketName, false)
	if minioErr != nil {
		return
	}

	// defaults to private
	policy := ""
	if bucket.Spec.Policy.Custom != "" {
		policy = bucket.Spec.Policy.Custom
	} else {
		if bucket.Spec.Policy.Preset == corev1alpha1.BucketCDNPreset {
			var tmpl *template.Template
			tmpl, minioErr = template.New("policy").Parse(cdnPolicy)
			if minioErr != nil {
				return
			}
			buff := bytes.Buffer{}
			minioErr = tmpl.Execute(&buff, bucket.Spec.BucketName)
			if minioErr != nil {
				return
			}
			policy = buff.String()
		}
	}
	minioErr = mc.SetBucketPolicy(ctx, bucket.Spec.BucketName, policy)
	if minioErr != nil {
		return
	}

	if mc.Provider.BucketDNS != nil {
		var ingress networkingv1.Ingress
		ingress.Name = fmt.Sprintf("bucket-%s", bucket.Spec.BucketName)
		ingress.Namespace = mc.Provider.BucketDNS.Namespace
		_, err = ctrl.CreateOrUpdate(ctx, r.Client, &ingress, func() error {
			ingress.Annotations = map[string]string{
				"nginx.ingress.kubernetes.io/configuration-snippet": "chunked_transfer_encoding off;",
				"nginx.ingress.kubernetes.io/proxy-body-size":       "0",
				"nginx.ingress.kubernetes.io/proxy-ssl-verify":      "off",
				"nginx.ingress.kubernetes.io/rewrite-target":        "/",
				"nginx.ingress.kubernetes.io/server-snippet":        "client_max_body_size 0;",
			}
			for k, v := range mc.Provider.BucketDNS.Annotations {
				ingress.Annotations[k] = v
			}
			pathType := networkingv1.PathTypePrefix
			ingress.Spec.Rules = []networkingv1.IngressRule{
				{
					Host: bucketEndpoint,
					IngressRuleValue: networkingv1.IngressRuleValue{
						HTTP: &networkingv1.HTTPIngressRuleValue{
							Paths: []networkingv1.HTTPIngressPath{
								{
									PathType: &pathType,
									Path:     "/",
									Backend: networkingv1.IngressBackend{
										Service: &mc.Provider.BucketDNS.ServiceBackend,
									},
								},
							},
						},
					},
				},
			}
			if !mc.Provider.Insecure {
				ingress.Spec.TLS = []networkingv1.IngressTLS{
					{
						Hosts:      []string{bucketEndpoint},
						SecretName: fmt.Sprintf("%s-tls", ingress.Name),
					},
				}
			}
			return nil
		})
	}

	// minioErr = fmt.Errorf("mean error")
	conditions.Delete(&bucket, meta.ReconcilingCondition)
	return
}

// SetupWithManager sets up the controller with the Manager.
func (r *BucketReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&corev1alpha1.Bucket{}).
		Complete(r)
}

func (r *BucketReconciler) deleteExternalResources(ctx context.Context, bucket *corev1alpha1.Bucket) error {
	mc, err := internal.GetMinioClient(ctx, r.Client, r.Config, bucket)
	if err != nil {
		return err
	}
	err = mc.RemoveBucketWithOptions(ctx, bucket.Spec.BucketName, minio.RemoveBucketOptions{
		ForceDelete: true,
	})
	if err != nil {
		return err
	}
	err = mc.Admin.RemoveUser(ctx, bucket.Spec.BucketName)
	return err
}
