package core

import (
	"context"
	"fmt"
	"strings"

	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/runtime"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/controller/controllerutil"
	"sigs.k8s.io/controller-runtime/pkg/log"

	configv1alpha1 "libre.sh/controller/apis/config/v1alpha1"
	corev1alpha1 "libre.sh/controller/apis/core/v1alpha1"
	"libre.sh/controller/internal"
)

// MailReconciler reconciles a Mail object
type MailReconciler struct {
	client.Client
	Scheme *runtime.Scheme
	Config *configv1alpha1.LibreConfig
}

//+kubebuilder:rbac:groups=core.libre.sh,resources=mails,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=core.libre.sh,resources=mails/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=core.libre.sh,resources=mails/finalizers,verbs=update

// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the Mail object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.13.0/pkg/reconcile
func (r *MailReconciler) Reconcile(ctx context.Context, req ctrl.Request) (result ctrl.Result, err error) {
	_ = log.FromContext(ctx)

	var mail corev1alpha1.Mail
	err = r.Client.Get(ctx, req.NamespacedName, &mail)
	if err != nil {
		return ctrl.Result{}, client.IgnoreNotFound(err)
	}

	provider := r.Config.GetViMbAdminProvider("default")
	if provider == nil {
		err = fmt.Errorf("provider %s not found", "default")
		return
	}

	vc := internal.NewViMbAdminClient(provider)
	err = vc.Login()
	if err != nil {
		return
	}
	at := strings.LastIndex(mail.Spec.Address, "@")
	if at < 0 {
		err = fmt.Errorf("%s is an invalid email address\n", mail.Spec.Address)
		return
	}
	username, domain := mail.Spec.Address[:at], mail.Spec.Address[at+1:]

	domainID, err := vc.GetDomainID(domain)
	if err != nil {
		return
	}

	var secret corev1.Secret
	secret.Name = mail.GetSecretName()
	secret.Namespace = mail.Namespace
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &secret, func() error {
		password := secret.Data["password"]
		if len(password) == 0 {
			p, err := vc.CreateMailbox(domainID, username)
			if err != nil {
				return err
			}
			password = []byte(p)
		}

		secret.Data = map[string][]byte{
			"username":          []byte(mail.Spec.Address),
			"password":          password,
			"host":              []byte(provider.Endpoint),
			"port":              []byte("587"),
			"mail_from_address": []byte(username),
			"from_email":        []byte(mail.Spec.Address),
		}
		return controllerutil.SetOwnerReference(&mail, &secret, r.Scheme)
	})
	if err != nil {
		return
	}

	return ctrl.Result{}, nil
}

// SetupWithManager sets up the controller with the Manager.
func (r *MailReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&corev1alpha1.Mail{}).
		Complete(r)
}
