package core

import (
	"context"
	"fmt"
	"sort"

	"github.com/fluxcd/pkg/apis/meta"
	"github.com/fluxcd/pkg/runtime/conditions"
	appv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"
	rbacv1 "k8s.io/api/rbac/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/types"
	"k8s.io/apimachinery/pkg/util/intstr"
	configv1alpha1 "libre.sh/controller/apis/config/v1alpha1"
	corev1alpha1 "libre.sh/controller/apis/core/v1alpha1"
	"libre.sh/controller/internal"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/builder"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/controller/controllerutil"
	"sigs.k8s.io/controller-runtime/pkg/handler"
	"sigs.k8s.io/controller-runtime/pkg/log"
	"sigs.k8s.io/controller-runtime/pkg/predicate"
	"sigs.k8s.io/controller-runtime/pkg/reconcile"
	"sigs.k8s.io/controller-runtime/pkg/source"
)

// RedisReconciler reconciles a Redis object
type RedisReconciler struct {
	client.Client
	Scheme *runtime.Scheme
	Config *configv1alpha1.LibreConfig
}

//+kubebuilder:rbac:groups=core.libre.sh,resources=redis,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=core.libre.sh,resources=redis/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=core.libre.sh,resources=redis/finalizers,verbs=update

// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the Redis object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.12.2/pkg/reconcile
func (r *RedisReconciler) Reconcile(ctx context.Context, req ctrl.Request) (result ctrl.Result, err error) {
	log := log.FromContext(ctx)

	var redis corev1alpha1.Redis
	err = r.Client.Get(ctx, req.NamespacedName, &redis)
	if err != nil {
		return ctrl.Result{}, client.IgnoreNotFound(err)
	}

	if redis.Spec.Suspend {
		return ctrl.Result{}, nil
	}

	log.Info("Reconciling")

	defer internal.UpdateStatus(ctx, r.Client, &redis, &err)

	if redis.Generation != conditions.GetObservedGeneration(&redis, meta.ReadyCondition) {
		conditions.MarkReconciling(&redis, meta.ProgressingReason, "Reconciling in progress")
		err = r.Client.Status().Update(ctx, &redis)
		if err != nil {
			return
		}
	}

	labels := map[string]string{
		"app.kubernetes.io/name":       "keydb",
		"app.kubernetes.io/instance":   redis.Name,
		"app.kubernetes.io/managed-by": "libre.sh",
	}

	serviceAccount := corev1.ServiceAccount{}
	serviceAccount.Name = "keydb"
	serviceAccount.Namespace = redis.Namespace
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &serviceAccount, func() error {
		return nil
	})
	if err != nil {
		return
	}

	role := rbacv1.Role{}
	role.Name = "keydb"
	role.Namespace = redis.Namespace
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &role, func() error {
		role.Rules = []rbacv1.PolicyRule{
			{
				APIGroups: []string{""},
				Resources: []string{"pods"},
				Verbs:     []string{"get", "list", "watch"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"endpoints"},
				Verbs:     []string{"get"},
			},
		}
		return nil
	})
	if err != nil {
		return
	}

	roleBinding := rbacv1.RoleBinding{}
	roleBinding.Name = "keydb"
	roleBinding.Namespace = redis.Namespace
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &roleBinding, func() error {
		roleBinding.RoleRef = rbacv1.RoleRef{
			APIGroup: role.GroupVersionKind().Group,
			Kind:     role.Kind,
			Name:     role.Name,
		}
		roleBinding.Subjects = []rbacv1.Subject{
			{
				Kind: serviceAccount.GroupVersionKind().Kind,
				Name: serviceAccount.Name,
			},
		}
		return nil
	})
	if err != nil {
		return
	}

	service := corev1.Service{}
	service.Name = fmt.Sprintf("%s-keydb", redis.Name)
	service.SetNamespace(redis.Namespace)
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &service, func() error {
		service.Labels = labels
		service.Spec = corev1.ServiceSpec{
			Ports: []corev1.ServicePort{
				{
					Port:       6379,
					TargetPort: intstr.FromInt(6379),
				},
			},
		}
		return controllerutil.SetOwnerReference(&redis, &service, r.Scheme)
	})
	if err != nil {
		return
	}

	endpoints := corev1.Endpoints{}
	endpoints.Name = fmt.Sprintf("%s-keydb", redis.Name)
	endpoints.SetNamespace(redis.Namespace)
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &endpoints, func() error {
		ip := getEndpointsIP(&endpoints)
		mainPod, err := r.getMainPod(ctx, &redis, ip)
		if err != nil {
			return err
		}
		endpoints.Labels = labels
		endpoints.Subsets = []corev1.EndpointSubset{}
		if mainPod != nil {
			endpoints.Subsets = []corev1.EndpointSubset{
				{
					Addresses: []corev1.EndpointAddress{
						{
							IP:       mainPod.Status.PodIP,
							Hostname: mainPod.Name,
							TargetRef: &corev1.ObjectReference{
								Kind:      mainPod.Kind,
								Namespace: mainPod.Namespace,
								Name:      mainPod.Name,
								UID:       mainPod.UID,
							},
						},
					},
					Ports: []corev1.EndpointPort{
						{
							Port:     6379,
							Protocol: corev1.ProtocolTCP,
						},
					},
				},
			}
		}
		return nil
	})
	if err != nil {
		return
	}

	sts := appv1.StatefulSet{}
	sts.Name = fmt.Sprintf("keydb-%s", redis.Name)
	sts.Namespace = redis.Namespace
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &sts, func() error {
		sts.Labels = labels
		replicas := int32(2)
		sts.Spec.Replicas = &replicas
		sts.Spec.ServiceName = service.Name
		sts.Spec.Selector = &metav1.LabelSelector{
			MatchLabels: labels,
		}
		probe := corev1.Probe{
			ProbeHandler: corev1.ProbeHandler{
				HTTPGet: &corev1.HTTPGetAction{
					Path: "/healthz",
					Port: intstr.FromInt(80),
				},
			},
		}
		sts.Spec.Template = corev1.PodTemplateSpec{
			ObjectMeta: metav1.ObjectMeta{
				Labels: labels,
			},
			Spec: corev1.PodSpec{
				ServiceAccountName: serviceAccount.GetName(),
				Containers: []corev1.Container{
					{
						Name:  "keydb",
						Image: "eqalpha/keydb:alpine_x86_64_v6.3.1",
						Args:  []string{"--active-replica", "yes", "--multi-master", "yes"},
						Lifecycle: &corev1.Lifecycle{
							PreStop: &corev1.LifecycleHandler{
								HTTPGet: &corev1.HTTPGetAction{
									Path: "/graceful-shutdown/keydb",
									Port: intstr.FromInt(80),
								},
							},
						},
					},
					{
						Name:  "manager",
						Image: "libresh/keydb-manager:v0.0.2",
						Env: []corev1.EnvVar{
							{
								Name: "NAMESPACE",
								ValueFrom: &corev1.EnvVarSource{
									FieldRef: &corev1.ObjectFieldSelector{
										FieldPath: "metadata.namespace",
									},
								},
							},
							{
								Name:  "INSTANCE",
								Value: redis.Name,
							},
						},
						Lifecycle: &corev1.Lifecycle{
							PreStop: &corev1.LifecycleHandler{
								HTTPGet: &corev1.HTTPGetAction{
									Path: "/graceful-shutdown/manager",
									Port: intstr.FromInt(80),
								},
							},
						},
						ReadinessProbe: &probe,
						LivenessProbe:  &probe,
						StartupProbe: &corev1.Probe{
							PeriodSeconds:    1,
							FailureThreshold: 30,
							ProbeHandler:     probe.ProbeHandler,
						},
					},
				},
			},
		}
		return controllerutil.SetControllerReference(&redis, &sts, r.Scheme)
	})
	if err != nil {
		return
	}

	var secret corev1.Secret
	secret.Name = fmt.Sprintf("%s.redis.libre.sh", redis.Name)
	secret.Namespace = redis.Namespace
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &secret, func() error {
		secret.StringData = map[string]string{
			"endpoint": service.Name,
			"port":     "6379",
		}
		return controllerutil.SetOwnerReference(&redis, &secret, r.Scheme)
	})
	if err != nil {
		return
	}

	conditions.Delete(&redis, meta.ReconcilingCondition)
	conditions.Delete(&redis, meta.StalledCondition)

	return
}

// SetupWithManager sets up the controller with the Manager.
func (r *RedisReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&corev1alpha1.Redis{}).
		Watches(
			&source.Kind{Type: &corev1.Pod{}},
			handler.EnqueueRequestsFromMapFunc(r.findObjectsForPod),
			builder.WithPredicates(predicate.ResourceVersionChangedPredicate{}),
		).
		Complete(r)
}

func (r *RedisReconciler) findObjectsForPod(pod client.Object) []reconcile.Request {
	requests := []reconcile.Request{}
	labels := pod.GetLabels()
	if labels["app.kubernetes.io/managed-by"] == "libre.sh" &&
		labels["app.kubernetes.io/name"] == "keydb" &&
		labels["app.kubernetes.io/instance"] != "" {
		requests = append(requests, reconcile.Request{
			NamespacedName: types.NamespacedName{
				Name:      labels["app.kubernetes.io/instance"],
				Namespace: pod.GetNamespace(),
			},
		})
	}
	return requests
}

func (r *RedisReconciler) getMainPod(ctx context.Context, redis *corev1alpha1.Redis, ip string) (*corev1.Pod, error) {
	pods := corev1.PodList{}
	err := r.List(ctx, &pods, client.InNamespace(redis.Namespace), client.MatchingLabels{
		"app.kubernetes.io/name":       "keydb",
		"app.kubernetes.io/instance":   redis.Name,
		"app.kubernetes.io/managed-by": "libre.sh",
	})
	if err != nil {
		return nil, err
	}
	readyPods := []corev1.Pod{}
	for _, pod := range pods.Items {
		if !pod.DeletionTimestamp.IsZero() {
			continue
		}
		for _, condition := range pod.Status.Conditions {
			if condition.Type == "Ready" && condition.Status == "True" {
				readyPods = append(readyPods, pod)
			}
		}
	}
	if ip != "" {
		for _, pod := range readyPods {
			if pod.Status.PodIP == ip {
				return &pod, nil
			}
		}
	}
	sort.Slice(readyPods, func(i, j int) bool {
		return readyPods[i].Name < readyPods[j].Name
	})
	if len(readyPods) > 0 {
		return &readyPods[0], nil
	}
	return nil, nil
}

func getEndpointsIP(endpoints *corev1.Endpoints) string {
	for _, subset := range endpoints.Subsets {
		for _, addresse := range subset.Addresses {
			return addresse.IP
		}
	}
	return ""
}
