package internal

import (
	"errors"
	"fmt"
	"net/http"
	"net/http/cookiejar"
	"net/url"
	"strings"

	"github.com/PuerkitoBio/goquery"
	"k8s.io/apimachinery/pkg/util/rand"
	configv1alpha1 "libre.sh/controller/apis/config/v1alpha1"
)

type ViMbAdminClient struct {
	provider *configv1alpha1.Provider
	cookies  []*http.Cookie
	client   *http.Client
}

func NewViMbAdminClient(provider *configv1alpha1.Provider) *ViMbAdminClient {
	jar, _ := cookiejar.New(nil)
	client := &http.Client{
		Jar: jar,
	}
	return &ViMbAdminClient{
		provider: provider,
		client:   client,
	}
}

func (vc *ViMbAdminClient) url() url.URL {
	u := url.URL{}
	u.Host = vc.provider.Endpoint
	u.Scheme = "https"
	if vc.provider.Insecure {
		u.Scheme = "http"
	}
	return u
}

func (vc *ViMbAdminClient) Login() error {
	u := vc.url()
	u.Path = "/auth/login"

	params := url.Values{}
	params.Add("username", vc.provider.Username)
	params.Add("password", vc.provider.Password)
	params.Add("login", `Log In`)
	params.Add("rememberme", `0`)
	body := strings.NewReader(params.Encode())

	req, err := http.NewRequest("POST", u.String(), body)
	if err != nil {
		return err
	}
	req.Header.Set("Content-Type", "application/x-www-form-urlencoded")

	resp, err := vc.client.Do(req)
	if err != nil {
		return err
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		return fmt.Errorf("failed to login : %s", resp.Status)
	}

	return nil
}

func (vc *ViMbAdminClient) GetDomainID(domain string) (string, error) {
	if domain == "" {
		return "", errors.New("domain can't be empty")
	}

	u := vc.url()
	u.Path = "/domain"

	req, err := http.NewRequest("GET", u.String(), nil)
	if err != nil {
		return "", err
	}

	resp, err := vc.client.Do(req)
	if err != nil {
		return "", err
	}

	if resp.StatusCode != http.StatusOK {
		return "", fmt.Errorf("failed to check domain : %s", resp.Status)
	}
	defer resp.Body.Close()

	doc, err := goquery.NewDocumentFromReader(resp.Body)
	if err != nil {
		return "", err
	}

	var id string
	doc.Find("tr").Each(func(i int, s *goquery.Selection) {
		if domain == s.Find("td").First().Text() {
			attr, _ := s.Attr("id")
			id = strings.ReplaceAll(attr, "domain_", "")
		}
	})

	return id, nil
}

func (vc *ViMbAdminClient) CreateMailbox(domainID string, username string) (string, error) {
	u := vc.url()
	u.Path = "/mailbox/add"

	password := rand.String(32)

	params := url.Values{}
	params.Add("local_part", username)
	params.Add("password", password)
	params.Add("domain", domainID)
	body := strings.NewReader(params.Encode())

	req, err := http.NewRequest("POST", u.String(), body)
	if err != nil {
		return "", err
	}
	req.Header.Set("Content-Type", "application/x-www-form-urlencoded")

	resp, err := vc.client.Do(req)
	if err != nil {
		return "", err
	}

	if resp.StatusCode != http.StatusOK {
		return "", fmt.Errorf("failed to create mailbox %s in domain %s : %s", username, domainID, resp.Status)
	}
	defer resp.Body.Close()

	return password, nil
}
